#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import itertools
import tempfile
import warnings

import _util
import numpy as np
import pkg_resources
import pytest
import sklearn.utils.estimator_checks
import torch
import torch.nn as nn

import cebra.data as cebra_data
import cebra.helper
import cebra.integrations.sklearn.cebra as cebra_sklearn_cebra
import cebra.integrations.sklearn.dataset as cebra_sklearn_dataset
import cebra.integrations.sklearn.utils as cebra_sklearn_utils
import cebra.models
import cebra.models.model
from cebra.models import parametrize

if torch.cuda.is_available():
    _DEVICES = "cpu", "cuda"
else:
    _DEVICES = ("cpu",)


def test_imports():
    import cebra

    assert hasattr(cebra, "CEBRA")


def test_sklearn_dataset():
    X = np.zeros((100, 5), dtype="float32")
    yc = np.zeros((100, 5), dtype="float32")
    yd = np.zeros((100,), dtype="int")

    # cannot create datasets with more than one 1D discrete index
    with pytest.raises(ValueError):
        cebra_sklearn_dataset.SklearnDataset(X, (yd, yd), device="cpu")
    with pytest.raises(ValueError):
        cebra_sklearn_dataset.SklearnDataset(X, (np.stack([yd, yd], axis=1),),
                                             device="cpu")

    # need to pass iterable as type ...
    with pytest.raises(TypeError):
        cebra_sklearn_dataset.SklearnDataset(X, yc, device="cpu")
    with pytest.raises(TypeError):
        cebra_sklearn_dataset.SklearnDataset(X, yd, device="cpu")
    with pytest.raises(TypeError):
        cebra_sklearn_dataset.SklearnDataset(X, [yd], device="cpu")
    with pytest.raises(TypeError):
        cebra_sklearn_dataset.SklearnDataset(X, [[yd]], device="cpu")
    with pytest.raises(TypeError):
        cebra_sklearn_dataset.SklearnDataset(X, [[[yd]]], device="cpu")
    with pytest.raises(ValueError):
        cebra_sklearn_dataset.SklearnDataset(X, ((yd,),), device="cpu")

    # ... but any iterable needs to work
    arg = ()
    cebra_sklearn_dataset.SklearnDataset(X, arg, device="cpu")
    arg = (yd,)
    cebra_sklearn_dataset.SklearnDataset(X, arg, device="cpu")
    arg = (yc,)
    cebra_sklearn_dataset.SklearnDataset(X, arg, device="cpu")
    arg = (yc, yc, yd)
    cebra_sklearn_dataset.SklearnDataset(X, arg, device="cpu")

    # checking data input and indexing ops
    for labels in [(), (yc,), (yd,), (yc, yd)]:
        data = cebra_sklearn_dataset.SklearnDataset(X, labels, device="cpu")
        assert data.input_dimension == X.shape[1]
        assert len(data) == len(X)
        assert data[torch.arange(10)].shape == X[:10, ..., np.newaxis].shape
        assert torch.allclose(data[torch.arange(10)],
                              torch.from_numpy(X[:10, ..., np.newaxis]))

    # check indexing
    for is_cont, is_disc in [[True, False]] * 2:
        arg = []
        if is_cont:
            arg.append(yc)
        if is_disc:
            arg.append(yd)
        arg = tuple(arg)
        data = cebra_sklearn_dataset.SklearnDataset(X, arg, device="cpu")
        assert (data.continuous_index is not None) == is_cont
        assert (data.discrete_index is not None) == is_disc

    # multisession
    num_sessions = 3

    # continuous
    sessions = []
    for i in range(num_sessions):
        sessions.append(cebra_sklearn_dataset.SklearnDataset(X, (yc,)))
    data = cebra_data.datasets.DatasetCollection(*sessions)
    assert data.num_sessions == num_sessions
    for i in range(num_sessions):
        assert data.get_input_dimension(i) == X.shape[1]
        assert len(data.get_session(i)) == len(X)

    with pytest.raises(ValueError):
        cebra_data.datasets.DatasetCollection(())
    with pytest.raises(ValueError):
        cebra_data.datasets.DatasetCollection(sessions)

    # discrete
    sessions = []
    for i in range(num_sessions):
        sessions.append(cebra_sklearn_dataset.SklearnDataset(X, (yd,)))
    data = cebra_data.datasets.DatasetCollection(*sessions)
    assert data.num_sessions == num_sessions
    for i in range(num_sessions):
        assert data.get_input_dimension(i) == X.shape[1]
        assert len(data.get_session(i)) == len(X)


@pytest.mark.parametrize("int_type", [np.uint8, np.int8, np.int32])
@pytest.mark.parametrize("float_type", [np.float16, np.float32, np.float64])
def test_sklearn_dataset_type_index(int_type, float_type):
    N = 100
    X = np.random.uniform(0, 1, (N * 2, 2))
    y = np.concatenate([np.zeros(N), np.ones(N)])

    # integer type
    y = y.astype(int_type)
    _, _, loader, _ = cebra.CEBRA(batch_size=512)._prepare_fit(X, y)
    assert loader.dataset.discrete_index is not None
    assert loader.dataset.continuous_index is None

    # floating type
    y = y.astype(float_type)
    _, _, loader, _ = cebra.CEBRA(batch_size=512)._prepare_fit(X, y)
    assert loader.dataset.continuous_index is not None
    assert loader.dataset.discrete_index is None


@_util.parametrize_slow(
    arg_names="is_cont,is_disc,is_full,is_multi,is_hybrid",
    fast_arguments=list(
        itertools.islice(itertools.product(*[[False, True]] * 5), 1)),
    slow_arguments=list(itertools.product(*[[False, True]] * 5)),
)
def test_init_loader(is_cont, is_disc, is_full, is_multi, is_hybrid):
    if is_multi and is_cont:
        # TODO(celia): change to a MultiDemoDataset class when discrete/mixed index implemented
        class __Dataset(cebra.datasets.MultiContinuous):
            neural = torch.zeros((50, 10), dtype=torch.float)
            continuous_index = torch.zeros((50, 10), dtype=torch.float)
    elif is_multi and is_disc:

        class __Dataset(cebra.datasets.MultiDiscrete):
            neural = torch.zeros((50, 10), dtype=torch.float)
            discrete_index = torch.zeros((50,), dtype=torch.int)
    else:

        class __Dataset(cebra.datasets.DemoDataset):
            neural = torch.zeros((50, 10), dtype=torch.float)
            continuous_index = torch.zeros((50, 10), dtype=torch.float)
            discrete_index = torch.zeros((50,), dtype=torch.int)

    shared_kwargs = dict(num_steps=5, dataset=__Dataset())
    extra_kwargs = dict(batch_size=512, time_offsets=10, delta=0.01)

    try:
        loader, solver = cebra_sklearn_cebra._init_loader(
            is_cont=is_cont,
            is_disc=is_disc,
            is_full=is_full,
            is_multi=is_multi,
            is_hybrid=is_hybrid,
            shared_kwargs=shared_kwargs,
            extra_kwargs=extra_kwargs,
        )
        assert isinstance(loader, cebra.data.Loader)
        assert isinstance(solver, str)
    except Exception as e:
        with pytest.raises((NotImplementedError, ValueError)):
            raise e


def iterate_models():
    # architecture checks
    for model_architecture, device, distance in itertools.product(
            cebra_sklearn_cebra.CEBRA.supported_model_architectures(),
            _DEVICES,
        ["euclidean", "cosine"],
    ):
        yield cebra_sklearn_cebra.CEBRA(
            model_architecture=model_architecture,
            pad_before_transform=
            True,  # NOTE(stes) needs to be true, otherwise not sklearn compatible.
            device=device,
            distance=distance,
            time_offsets=5,
            max_iterations=5,
            batch_size=10,
        )

    # parameter checks
    for (
            model_architecture,
            device,
            distance,
            temperature_mode,
            min_temperature,
            temperature,
    ) in itertools.product(
        [
            "offset10-model", "offset10-model-mse", "offset1-model",
            "resample-model"
        ],
            _DEVICES,
        ["euclidean", "cosine"],
        ["auto", "constant"],
        [None, 0.1],
        [0.1, 1.0],
    ):
        yield cebra_sklearn_cebra.CEBRA(
            model_architecture=model_architecture,
            pad_before_transform=
            True,  # NOTE(stes) needs to be true, otherwise not sklearn compatible.
            device=device,
            distance=distance,
            min_temperature=min_temperature,
            temperature_mode=temperature_mode,
            time_offsets=5,
            max_iterations=5,
            batch_size=10,
        )


@_util.parametrize_with_checks_slow(
    fast_arguments=list(itertools.islice(iterate_models(), 1)),
    slow_arguments=list(iterate_models()),
)
def test_api(estimator, check):
    num_retries = 1
    if check.func == sklearn.utils.estimator_checks.check_fit_idempotent:
        pytest.skip("CEBRA is non-deterministic.")
    if (check.func == sklearn.utils.estimator_checks.
            check_methods_sample_order_invariance):
        num_retries = 1000
    # if estimator.model_architecture == 'offset5-model':
    #    if check.func == sklearn.utils.estimator_checks.check_methods_sample_order_invariance:
    #        pytest.skip(
    #            "Output of fully convolutional models is not permutation invariant."
    #        )
    #    if check.func == sklearn.utils.estimator_checks.check_methods_subset_invariance:
    #        pytest.skip(
    #            "Output of fully convolutional models is not subset invariant.")
    if len(estimator._compute_offset()) > 10:
        pytest.skip(f"Model architecture {estimator.model_architecture} "
                    f"requires longer input sizes than 20 samples.")

    exception = None
    num_successful = 0
    total_runs = 0
    for _ in range(num_retries):
        total_runs += 1
        try:
            check(estimator)
            num_successful += 1
        except AssertionError as e:
            exception = e
        if total_runs > num_retries // 10:
            if num_successful > 0:
                break
    if num_successful == 0:
        raise exception
    if exception is not None:
        warnings.warn(
            UserWarning(
                f"At least one repeat of the test raised an assertion error. "
                f"Test has a success probability of {num_successful}/{total_runs} = {100. * num_successful/total_runs:.2f}%."
            ))


@_util.parametrize_slow(
    arg_names="model_architecture,device",
    fast_arguments=list(
        itertools.islice(
            itertools.product(
                cebra_sklearn_cebra.CEBRA.supported_model_architectures(),
                _DEVICES),
            2,
        )),
    slow_arguments=list(
        itertools.product(
            cebra_sklearn_cebra.CEBRA.supported_model_architectures(),
            _DEVICES)),
)
def test_sklearn(model_architecture, device):
    output_dimension = 4
    cebra_model = cebra_sklearn_cebra.CEBRA(
        model_architecture=model_architecture,
        time_offsets=10,
        learning_rate=3e-4,
        max_iterations=5,
        device=device,
        output_dimension=output_dimension,
        batch_size=42,
        verbose=True,
    )

    # example dataset
    X = np.random.uniform(0, 1, (1000, 50))
    X_s2 = np.random.uniform(0, 1, (800, 30))
    X_s3 = np.random.uniform(0, 1, (1000, 30))
    y_c1 = np.random.uniform(0, 1, (1000, 5))
    y_c1_s2 = np.random.uniform(0, 1, (800, 5))
    y_c2 = np.random.uniform(0, 1, (1000, 2))
    y_d = np.random.randint(0, 10, (1000,))
    y_d_s2 = np.random.randint(0, 10, (800,))

    # time contrastive
    cebra_model.fit(X)
    assert cebra_model.num_sessions is None
    embedding = cebra_model.transform(X)
    assert isinstance(embedding, np.ndarray)

    # continuous behavior contrastive
    cebra_model.fit(X, y_c1, y_c2)
    assert cebra_model.num_sessions is None

    embedding = cebra_model.transform(X)
    assert isinstance(embedding, np.ndarray)
    embedding = cebra_model.transform(torch.Tensor(X))
    assert isinstance(embedding, np.ndarray)
    embedding = cebra_model.transform(X, session_id=0)
    assert isinstance(embedding, np.ndarray)

    with pytest.raises(RuntimeError, match="Invalid.*session_id"):
        embedding = cebra_model.transform(X, session_id=2)
    with pytest.raises(ValueError, match="Invalid.*labels"):
        cebra_model.fit(X, [y_c1, y_c1_s2])
    with pytest.raises(ValueError, match="Invalid.*samples"):
        cebra_model.fit(X_s3, y_c1_s2)

    # tensor input
    cebra_model.fit(torch.Tensor(X), torch.Tensor(y_c1), torch.Tensor(y_c2))

    # discrete behavior contrastive
    cebra_model.fit(X, y_d)
    embedding = cebra_model.transform(X)
    assert isinstance(embedding, np.ndarray)

    # mixed
    cebra_model.fit(X, y_c1, y_c2, y_d)
    embedding = cebra_model.transform(X)
    assert isinstance(embedding, np.ndarray)

    # multi-session discrete behavior contrastive
    cebra_model.fit([X, X_s2], [y_d, y_d_s2])
    assert cebra_model.num_sessions == 2

    embedding = cebra_model.transform(X, session_id=0)
    assert isinstance(embedding, np.ndarray)
    embedding = cebra_model.transform(torch.Tensor(X), session_id=0)
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (X.shape[0], output_dimension)
    embedding = cebra_model.transform(X_s2, session_id=1)
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (X_s2.shape[0], output_dimension)

    with pytest.raises(ValueError, match="shape"):
        embedding = cebra_model.transform(X_s2, session_id=0)
    with pytest.raises(ValueError, match="shape"):
        embedding = cebra_model.transform(X, session_id=1)
    with pytest.raises(RuntimeError, match="No.*session_id"):
        embedding = cebra_model.transform(X)
    with pytest.raises(RuntimeError, match="Invalid.*session_id"):
        embedding = cebra_model.transform(X, session_id=2)
    with pytest.raises(RuntimeError, match="Invalid.*session_id"):
        embedding = cebra_model.transform(X, session_id=-1)

    # multi-session continuous behavior contrastive
    cebra_model.fit([X, X_s2], [y_c1, y_c1_s2])
    assert cebra_model.num_sessions == 2

    embedding = cebra_model.transform(X, session_id=0)
    assert isinstance(embedding, np.ndarray)
    embedding = cebra_model.transform(torch.Tensor(X), session_id=0)
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (X.shape[0], output_dimension)
    embedding = cebra_model.transform(X_s2, session_id=1)
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (X_s2.shape[0], output_dimension)

    with pytest.raises(ValueError, match="shape"):
        embedding = cebra_model.transform(X_s2, session_id=0)
    with pytest.raises(ValueError, match="shape"):
        embedding = cebra_model.transform(X, session_id=1)
    with pytest.raises(RuntimeError, match="No.*session_id"):
        embedding = cebra_model.transform(X)
    with pytest.raises(RuntimeError, match="Invalid.*session_id"):
        embedding = cebra_model.transform(X, session_id=2)
    with pytest.raises(RuntimeError, match="Invalid.*session_id"):
        embedding = cebra_model.transform(X, session_id=-1)

    # multi-session tensor inputs
    cebra_model.fit(
        [torch.Tensor(X), torch.Tensor(X_s2)],
        [torch.Tensor(y_c1), torch.Tensor(y_c1_s2)],
    )

    # multi-session discrete behavior contrastive, more than two sessions
    cebra_model.fit([X, X_s2, X], [y_d, y_d_s2, y_d])
    assert cebra_model.num_sessions == 3

    embedding = cebra_model.transform(X, session_id=0)
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (X.shape[0], output_dimension)
    embedding = cebra_model.transform(X_s2, session_id=1)
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (X_s2.shape[0], output_dimension)
    embedding = cebra_model.transform(X, session_id=2)
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (X.shape[0], output_dimension)

    with pytest.raises(ValueError, match="shape"):
        embedding = cebra_model.transform(X_s2, session_id=0)
    with pytest.raises(ValueError, match="shape"):
        embedding = cebra_model.transform(X_s2, session_id=2)
    with pytest.raises(ValueError, match="shape"):
        embedding = cebra_model.transform(X, session_id=1)
    with pytest.raises(RuntimeError, match="No.*session_id"):
        embedding = cebra_model.transform(X)
    with pytest.raises(RuntimeError, match="Invalid.*session_id"):
        embedding = cebra_model.transform(X, session_id=3)

    # multi-session continuous behavior contrastive, more than two sessions
    cebra_model.fit([X, X_s2, X], [y_c1, y_c1_s2, y_c1])
    assert cebra_model.num_sessions == 3

    embedding = cebra_model.transform(X, session_id=0)
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (X.shape[0], output_dimension)
    embedding = cebra_model.transform(X_s2, session_id=1)
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (X_s2.shape[0], output_dimension)
    embedding = cebra_model.transform(X, session_id=2)
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (X.shape[0], output_dimension)

    with pytest.raises(ValueError, match="shape"):
        embedding = cebra_model.transform(X_s2, session_id=0)
    with pytest.raises(ValueError, match="shape"):
        embedding = cebra_model.transform(X_s2, session_id=2)
    with pytest.raises(ValueError, match="shape"):
        embedding = cebra_model.transform(X, session_id=1)
    with pytest.raises(RuntimeError, match="No.*session_id"):
        embedding = cebra_model.transform(X)
    with pytest.raises(RuntimeError, match="Invalid.*session_id"):
        embedding = cebra_model.transform(X, session_id=3)

    with pytest.raises(RuntimeError, match="No.*label"):
        cebra_model.fit([X, X_s2])
    with pytest.raises(ValueError, match="Invalid.*sessions"):
        cebra_model.fit([X, X_s2], y_c1)
    with pytest.raises(ValueError, match="Invalid.*sessions"):
        cebra_model.fit(X, [y_c1, y_c2])
    with pytest.raises(ValueError, match="Invalid.*sessions"):
        cebra_model.fit([X, X, X_s2], [y_c1, y_c2])
    with pytest.raises(ValueError, match="Invalid.*sessions"):
        cebra_model.fit([X, X_s2], [y_c1, y_c1, y_c2])
    with pytest.raises(ValueError, match="Invalid.*samples"):
        cebra_model.fit([X, X_s2], [y_c1_s2, y_c1_s2])

    if cebra_model.pad_before_transform:
        assert embedding.shape == (len(X), 4)
    else:
        assert embedding.shape == (len(X) -
                                   len(cebra_model.model_.get_offset()) + 1, 4)

    for key in [
            "model", "optimizer", "loss", "decode", "criterion", "version",
            "log"
    ]:
        assert key in cebra_model.state_dict_, cebra_model.state_dict_.keys()


def test_adapt_model():
    adaptation_param_key = "net.O.weight", "net.0.bias"
    output_dimension = 4
    cebra_model = cebra_sklearn_cebra.CEBRA(
        model_architecture="offset1-model",
        time_offsets=10,
        learning_rate=3e-4,
        max_iterations=5,
        max_adapt_iterations=1,
        output_dimension=output_dimension,
        batch_size=42,
        verbose=True,
    )

    X = np.random.uniform(0, 1, (1000, 50))
    X_s2 = np.random.uniform(0, 1, (800, 30))

    cebra_model.fit(X)
    before_adapt = cebra_model.state_dict_["model"]
    cebra_model._adapt_model(X_s2)
    after_adapt = cebra_model.state_dict_["model"]

    assert before_adapt.keys() == after_adapt.keys()
    for key in before_adapt.keys():
        if key in adaptation_param_key:
            assert (before_adapt[key].shape
                    != after_adapt[key].shape) or not torch.allclose(
                        before_adapt[key], after_adapt[key])
        else:
            assert torch.allclose(before_adapt[key], after_adapt[key])


@_util.parametrize_slow(
    arg_names="model_architecture,device",
    fast_arguments=list(
        itertools.islice(
            itertools.product(
                cebra_sklearn_cebra.CEBRA.supported_model_architectures(),
                _DEVICES),
            2,
        )),
    slow_arguments=list(
        itertools.product(
            cebra_sklearn_cebra.CEBRA.supported_model_architectures(),
            _DEVICES)),
)
def test_partial_fit(model_architecture, device):
    max_iterations = 10
    partial_max_iterations = 5

    cebra_model = cebra_sklearn_cebra.CEBRA(
        model_architecture=model_architecture,
        time_offsets=10,
        learning_rate=3e-4,
        max_iterations=max_iterations,
        device=device,
        output_dimension=4,
        batch_size=42,
        verbose=True,
    )

    X = np.random.uniform(0, 1, (1000, 50))
    X_s2 = np.random.uniform(0, 1, (800, 30))
    y_c1 = np.random.uniform(0, 1, (1000, 5))
    y_c1_s2 = np.random.uniform(0, 1, (800, 5))

    # Single session training
    # Fit the model
    cebra_model.fit(X)
    assert len(cebra_model.state_dict_["loss"]) == max_iterations
    for k in cebra_model.state_dict_["log"].keys():
        assert len(cebra_model.state_dict_["log"][k]) == max_iterations

    # Assert that fitting the model again resets the fitting
    cebra_model.fit(X)
    assert len(cebra_model.state_dict_["loss"]) == max_iterations
    for k in cebra_model.state_dict_["log"].keys():
        assert len(cebra_model.state_dict_["log"][k]) == max_iterations

    cebra_partial_model = cebra_sklearn_cebra.CEBRA(
        model_architecture=model_architecture,
        time_offsets=10,
        learning_rate=3e-4,
        max_iterations=partial_max_iterations,
        device=device,
        output_dimension=4,
        batch_size=42,
        verbose=True,
    )

    # Partially fit the model
    cebra_partial_model.partial_fit(X)
    assert len(
        cebra_partial_model.state_dict_["loss"]) == partial_max_iterations
    for k in cebra_partial_model.state_dict_["log"].keys():
        assert len(
            cebra_partial_model.state_dict_["log"][k]) == partial_max_iterations

    # Assert that partially fitting the model again iterates over the previously fitted model
    cebra_partial_model.partial_fit(X)
    assert len(cebra_partial_model.state_dict_["loss"]) == max_iterations
    for k in cebra_partial_model.state_dict_["log"].keys():
        assert len(cebra_partial_model.state_dict_["log"][k]) == max_iterations

    # Multi session training
    # Fit the model
    cebra_model.fit([X, X_s2], [y_c1, y_c1_s2])
    assert len(cebra_model.state_dict_["loss"]) == max_iterations
    for k in cebra_model.state_dict_["log"].keys():
        assert len(cebra_model.state_dict_["log"][k]) == max_iterations

    # Assert that fitting the model again resets the fitting
    cebra_model.fit([X, X_s2], [y_c1, y_c1_s2])
    assert len(cebra_model.state_dict_["loss"]) == max_iterations
    for k in cebra_model.state_dict_["log"].keys():
        assert len(cebra_model.state_dict_["log"][k]) == max_iterations

    cebra_partial_model = cebra_sklearn_cebra.CEBRA(
        model_architecture=model_architecture,
        time_offsets=10,
        learning_rate=3e-4,
        max_iterations=partial_max_iterations,
        device=device,
        output_dimension=4,
        batch_size=42,
        verbose=True,
    )

    # Partially fit the model
    cebra_partial_model = cebra_sklearn_cebra.CEBRA(
        model_architecture=model_architecture,
        time_offsets=10,
        learning_rate=3e-4,
        max_iterations=partial_max_iterations,
        device=device,
        output_dimension=4,
        batch_size=42,
        verbose=True,
    )

    cebra_partial_model.partial_fit([X, X_s2], [y_c1, y_c1_s2])
    assert len(
        cebra_partial_model.state_dict_["loss"]) == partial_max_iterations
    for k in cebra_partial_model.state_dict_["log"].keys():
        assert len(
            cebra_partial_model.state_dict_["log"][k]) == partial_max_iterations

    # Assert that partially fitting the model again iterates over the previously fitted model
    cebra_partial_model.partial_fit([X, X_s2], [y_c1, y_c1_s2])
    assert len(cebra_partial_model.state_dict_["loss"]) == max_iterations
    for k in cebra_partial_model.state_dict_["log"].keys():
        assert len(cebra_partial_model.state_dict_["log"][k]) == max_iterations


@_util.parametrize_slow(
    arg_names="model_architecture,device",
    fast_arguments=list(
        itertools.islice(
            itertools.product(
                cebra_sklearn_cebra.CEBRA.supported_model_architectures(),
                _DEVICES),
            2,
        )),
    slow_arguments=list(
        itertools.product(
            cebra_sklearn_cebra.CEBRA.supported_model_architectures(),
            _DEVICES)),
)
def test_sklearn_adapt(model_architecture, device):
    num_hidden_units = 32
    cebra_model = cebra_sklearn_cebra.CEBRA(
        model_architecture=model_architecture,
        time_offsets=10,
        learning_rate=3e-4,
        max_iterations=5,
        max_adapt_iterations=1,
        device=device,
        output_dimension=4,
        num_hidden_units=num_hidden_units,
        batch_size=42,
        verbose=True,
    )

    # example dataset
    X = np.random.uniform(0, 1, (1000, 50))
    X_s2 = np.random.uniform(0, 1, (800, 30))
    X_s3 = np.random.uniform(0, 1, (1000, 30))
    y_c1 = np.random.uniform(0, 1, (1000, 5))
    y_c1_s2 = np.random.uniform(0, 1, (800, 5))
    y_c2 = np.random.uniform(0, 1, (1000, 2))
    y_c2_s2 = np.random.uniform(0, 1, (800, 2))
    y_d = np.random.randint(0, 10, (1000,))
    y_d_s2 = np.random.randint(0, 10, (800,))

    def check_first_layer_dim(model, X):
        params = model.state_dict_["model"]
        params_keys = list(params.keys())
        assert params[params_keys[0]].shape[1] == X.shape[1]

    cebra_model.fit(X)
    check_first_layer_dim(cebra_model, X)

    cebra_model.fit(X_s2, adapt=True)
    check_first_layer_dim(cebra_model, X_s2)
    embedding = cebra_model.transform(X_s2)
    assert isinstance(embedding, np.ndarray)

    cebra_model.fit(X, y_c1, y_c2, adapt=True)
    check_first_layer_dim(cebra_model, X)
    embedding = cebra_model.transform(X)
    assert isinstance(embedding, np.ndarray)
    embedding = cebra_model.transform(X, session_id=0)
    assert isinstance(embedding, np.ndarray)

    with pytest.raises(RuntimeError, match="Invalid.*session_id"):
        embedding = cebra_model.transform(X, session_id=2)
    with pytest.raises(ValueError, match="Invalid.*sessions"):
        cebra_model.fit(X_s2, [y_c1_s2, y_c2_s2], adapt=True)
    with pytest.raises(ValueError, match="Invalid.*samples"):
        cebra_model.fit(X_s3, y_c1_s2, adapt=True)

    cebra_model.fit(X_s2, y_d_s2, adapt=True)
    check_first_layer_dim(cebra_model, X_s2)
    embedding = cebra_model.transform(X_s2)
    assert isinstance(embedding, np.ndarray)

    cebra_model.fit(X, y_c1, y_c2, y_d, adapt=True)
    check_first_layer_dim(cebra_model, X)
    embedding = cebra_model.transform(X)
    assert isinstance(embedding, np.ndarray)

    with pytest.raises(NotImplementedError, match=".*multisession.*"):
        cebra_model.fit([X, X_s2], [y_c1, y_c1_s2], adapt=True)


@_util.parametrize_slow(
    arg_names="model_architecture,device",
    fast_arguments=list(
        itertools.islice(
            itertools.product(
                cebra_sklearn_cebra.CEBRA.supported_model_architectures(),
                _DEVICES),
            2,
        )),
    slow_arguments=list(
        itertools.product(
            cebra_sklearn_cebra.CEBRA.supported_model_architectures(),
            _DEVICES)),
)
def test_sklearn_adapt_multisession(model_architecture, device):
    num_hidden_units = 32
    cebra_model = cebra_sklearn_cebra.CEBRA(
        model_architecture=model_architecture,
        time_offsets=10,
        learning_rate=3e-4,
        max_iterations=5,
        max_adapt_iterations=1,
        device=device,
        output_dimension=4,
        num_hidden_units=num_hidden_units,
        batch_size=42,
        verbose=True,
    )

    # example dataset
    Xs = [np.random.uniform(0, 1, (1000, 50)) for i in range(3)]
    ys = [np.random.uniform(0, 1, (1000, 5)) for i in range(3)]

    X_new = np.random.uniform(0, 1, (1000, 50))
    y_new = np.random.uniform(0, 1, (1000, 5))

    cebra_model.fit(Xs, ys)

    with pytest.raises(NotImplementedError, match=".*multisession.*"):
        cebra_model.fit(X_new, y_new, adapt=True)


@_util.parametrize_slow(
    arg_names="model_architecture,device,pad_before_transform",
    fast_arguments=list(
        itertools.islice(
            itertools.product(
                cebra_sklearn_cebra.CEBRA.supported_model_architectures(),
                _DEVICES,
                [True, False],
            ),
            1,
        )),
    slow_arguments=list(
        itertools.product(
            cebra_sklearn_cebra.CEBRA.supported_model_architectures(),
            _DEVICES,
            [True, False],
        )),
)
def test_sklearn_full(model_architecture, device, pad_before_transform):
    cebra_model = cebra_sklearn_cebra.CEBRA(
        model_architecture=model_architecture,
        time_offsets=10,
        learning_rate=3e-4,
        max_iterations=5,
        device=device,
        pad_before_transform=pad_before_transform,
        output_dimension=4,
        batch_size=None,
        verbose=True,
    )

    # example dataset
    X = np.random.uniform(0, 1, (1000, 50))
    y_c1 = np.random.uniform(0, 1, (1000, 5))
    y_c2 = np.random.uniform(0, 1, (1000, 2))

    # time contrastive
    cebra_model.fit(X)
    embedding = cebra_model.transform(X)

    # continuous behavior contrastive
    cebra_model.fit(X, y_c1, y_c2)
    embedding = cebra_model.transform(X)

    # Check for https://github.com/stes/neural_cl/issues/153
    for target_type in [int, "float16", "int32", "int8", "uint8"]:
        embedding_disc_1 = cebra_model.transform(
            X.astype(target_type).astype("float32"))
        embedding_disc_2 = cebra_model.transform(X.astype(target_type))
        assert np.allclose(embedding_disc_1,
                           embedding_disc_2,
                           rtol=1e-2,
                           atol=1e-2), target_type
        assert embedding_disc_1.dtype == "float32"
        assert embedding_disc_2.dtype == "float32"

    assert isinstance(embedding, np.ndarray)
    if cebra_model.pad_before_transform:
        assert embedding.shape == (len(X), 4)
    else:
        assert embedding.shape == (len(X) -
                                   len(cebra_model.model_.get_offset()) + 1, 4)


@pytest.mark.parametrize("model_architecture,device",
                         [("resample-model", "cpu"),
                          ("resample5-model", "cpu")])
def test_sklearn_resampling_model(model_architecture, device):
    cebra_model = cebra_sklearn_cebra.CEBRA(
        model_architecture=model_architecture,
        time_offsets=10,
        learning_rate=3e-4,
        max_iterations=5,
        device=device,
        output_dimension=4,
        batch_size=128,
        verbose=True,
    )

    # example dataset
    X = torch.tensor(np.random.uniform(0, 1, (1000, 50)))
    y_c1 = torch.tensor(np.random.uniform(0, 1, (1000, 5)))

    cebra_model.fit(X, y_c1)
    output = cebra_model.transform(X)
    assert output.shape == (250, 4)


@pytest.mark.parametrize("model_architecture,device",
                         [("resample1-model", "cpu")])
def test_sklearn_resampling_model_not_yet_supported(model_architecture, device):
    cebra_model = cebra_sklearn_cebra.CEBRA(
        model_architecture=model_architecture, max_iterations=5)

    # example dataset
    X = torch.tensor(np.random.uniform(0, 1, (1000, 50)))
    y_c1 = torch.tensor(np.random.uniform(0, 1, (1000, 5)))

    with pytest.raises(ValueError):
        cebra_model.fit(X, y_c1)
        _ = cebra_model.transform(X)


def _iterate_actions():

    def do_nothing(model):
        return model

    def fit_singlesession_model(model):
        X = np.linspace(-1, 1, 1000)[:, None]
        model.fit(X)
        return model

    def fit_multisession_model(model):
        X = np.linspace(-1, 1, 1000)[:, None]
        model.fit([X, X], [X, X])
        return model

    return [do_nothing, fit_singlesession_model, fit_multisession_model]


def _assert_same_state_dict(first, second):
    assert first.keys() == second.keys()
    for key in first:
        if isinstance(first[key], torch.Tensor):
            assert torch.allclose(first[key], second[key]), key
        elif isinstance(first[key], dict):
            _assert_same_state_dict(first[key], second[key]), key
        else:
            assert first[key] == second[key]


def check_if_fit(model):
    """Check if a model was already fit.

    Args:
        model: The model to check.

    Returns:
        True if the model was already fit.
    """
    return hasattr(model, "n_features_")


def _assert_equal(original_model, loaded_model):
    assert original_model.get_params() == loaded_model.get_params()
    assert check_if_fit(loaded_model) == check_if_fit(original_model)

    if check_if_fit(loaded_model):
        _assert_same_state_dict(original_model.state_dict_,
                                loaded_model.state_dict_)
        X = np.random.normal(0, 1, (100, 1))

        if loaded_model.num_sessions is not None:
            assert np.allclose(loaded_model.transform(X, session_id=0),
                               original_model.transform(X, session_id=0))
        else:
            assert np.allclose(loaded_model.transform(X),
                               original_model.transform(X))


@parametrize(
    "parametrized-model-{output_dim}",
    output_dim=(5, 10),
)
class ParametrizedModelExample(cebra.models.model._OffsetModel):
    """CEBRA model with a single sample receptive field, without output normalization."""

    def __init__(
        self,
        num_neurons,
        num_units,
        num_output,
        output_dim,
        normalize=False,
    ):
        super().__init__(
            nn.Flatten(start_dim=1, end_dim=-1),
            nn.Linear(
                num_neurons,
                output_dim,
            ),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(0, 1)


@pytest.mark.parametrize("action", _iterate_actions())
@pytest.mark.parametrize("backend_save", ["torch", "sklearn"])
@pytest.mark.parametrize("backend_load", ["auto", "torch", "sklearn"])
@pytest.mark.parametrize("model_architecture",
                         ["offset1-model", "parametrized-model-5"])
@pytest.mark.parametrize("device", ["cpu"] +
                         ["cuda"] if torch.cuda.is_available() else [])
def test_save_and_load(action, backend_save, backend_load, model_architecture,
                       device):
    original_model = cebra_sklearn_cebra.CEBRA(
        model_architecture=model_architecture,
        max_iterations=5,
        batch_size=100,
        device=device)

    original_model = action(original_model)
    with tempfile.NamedTemporaryFile(mode="w+b", delete=True) as savefile:
        if not check_if_fit(original_model):
            with pytest.raises(ValueError):
                original_model.save(savefile.name, backend=backend_save)
        else:
            if "parametrized" in original_model.model_architecture and backend_save == "torch":
                with pytest.raises(AttributeError):
                    original_model.save(savefile.name, backend=backend_save)
            else:
                original_model.save(savefile.name, backend=backend_save)

                if (backend_load != "auto") and (backend_save != backend_load):
                    with pytest.raises(RuntimeError):
                        cebra_sklearn_cebra.CEBRA.load(savefile.name,
                                                       backend_load)
                else:
                    loaded_model = cebra_sklearn_cebra.CEBRA.load(
                        savefile.name, backend_load)
                    _assert_equal(original_model, loaded_model)
                    action(loaded_model)


def get_ordered_cuda_devices():
    available_devices = ['cuda']
    for i in range(torch.cuda.device_count()):
        available_devices.append(f'cuda:{i}')
    return available_devices


ordered_cuda_devices = get_ordered_cuda_devices() if torch.cuda.is_available(
) else []


def test_fit_after_moving_to_device():
    expected_device = 'cpu'
    expected_type = type(expected_device)

    X = np.random.uniform(0, 1, (10, 5))
    cebra_model = cebra_sklearn_cebra.CEBRA(model_architecture="offset1-model",
                                            max_iterations=5,
                                            device=expected_device)

    assert type(cebra_model.device) == expected_type
    assert cebra_model.device == expected_device

    cebra_model.partial_fit(X)
    assert type(cebra_model.device) == expected_type
    assert cebra_model.device == expected_device
    if hasattr(cebra_model, 'device_'):
        assert type(cebra_model.device_) == expected_type
        assert cebra_model.device_ == expected_device

    # Move the model to device using the to() method
    cebra_model.to('cpu')
    assert type(cebra_model.device) == expected_type
    assert cebra_model.device == expected_device
    if hasattr(cebra_model, 'device_'):
        assert type(cebra_model.device_) == expected_type
        assert cebra_model.device_ == expected_device

    cebra_model.partial_fit(X)
    assert type(cebra_model.device) == expected_type
    assert cebra_model.device == expected_device
    if hasattr(cebra_model, 'device_'):
        assert type(cebra_model.device_) == expected_type
        assert cebra_model.device_ == expected_device


@pytest.mark.parametrize("device", ['cpu'] + ordered_cuda_devices)
def test_move_cpu_to_cuda_device(device):

    if not torch.cuda.is_available():
        pytest.skip("CUDA not available")

    if device.startswith('cuda') and device not in ordered_cuda_devices:
        pytest.skip(f"Device ID {device} not available")

    X = np.random.uniform(0, 1, (10, 5))
    cebra_model = cebra_sklearn_cebra.CEBRA(model_architecture="offset1-model",
                                            max_iterations=5,
                                            device=device).fit(X)

    # Move the model to a different device
    new_device = 'cpu' if device.startswith('cuda') else 'cuda:0'
    cebra_model.to(new_device)

    assert cebra_model.device == new_device
    device_model = next(cebra_model.solver_.model.parameters()).device
    device_str = str(device_model)
    if device_model.type == 'cuda':
        device_str = f'cuda:{device_model.index}'
    assert device_str == new_device

    with tempfile.NamedTemporaryFile(mode="w+b", delete=True) as savefile:
        cebra_model.save(savefile.name)
        loaded_model = cebra_sklearn_cebra.CEBRA.load(savefile.name)

    assert cebra_model.device == loaded_model.device
    assert next(cebra_model.solver_.model.parameters()).device == next(
        loaded_model.solver_.model.parameters()).device


@pytest.mark.parametrize("device", ['cpu', "mps"])
def test_move_cpu_to_mps_device(device):

    if not cebra.helper._is_mps_availabe(torch):
        pytest.skip("MPS device is not available")

    X = np.random.uniform(0, 1, (10, 5))
    cebra_model = cebra_sklearn_cebra.CEBRA(model_architecture="offset1-model",
                                            max_iterations=5,
                                            device=device).fit(X)

    # Move the model to a different device
    new_device = 'cpu' if device == 'mps' else 'mps'
    cebra_model.to(new_device)

    assert cebra_model.device == new_device

    device_model = next(cebra_model.solver_.model.parameters()).device
    assert device_model.type == new_device

    with tempfile.NamedTemporaryFile(mode="w+b", delete=True) as savefile:
        cebra_model.save(savefile.name)
        loaded_model = cebra_sklearn_cebra.CEBRA.load(savefile.name)

    assert cebra_model.device == loaded_model.device
    assert next(cebra_model.solver_.model.parameters()).device == next(
        loaded_model.solver_.model.parameters()).device


ordered_cuda_devices = get_ordered_cuda_devices() if torch.cuda.is_available(
) else []


@pytest.mark.parametrize("device", ['mps'] + ordered_cuda_devices)
def test_move_mps_to_cuda_device(device):

    if (not torch.cuda.is_available()) or (
            not cebra.helper._is_mps_availabe(torch)):
        pytest.skip("CUDA or MPS not available")

    if device.startswith('cuda') and device not in ordered_cuda_devices:
        pytest.skip(f"Device ID {device} not available")

    X = np.random.uniform(0, 1, (10, 5))
    cebra_model = cebra_sklearn_cebra.CEBRA(model_architecture="offset1-model",
                                            max_iterations=5,
                                            device=device).fit(X)

    # Move the model to a different device
    new_device = 'mps' if device.startswith('cuda') else 'cuda:0'
    cebra_model.to(new_device)

    assert cebra_model.device == new_device
    device_model = next(cebra_model.solver_.model.parameters()).device
    device_str = str(device_model)
    if device_model.type == 'cuda':
        device_str = f'cuda:{device_model.index}'
    assert device_str == new_device

    with tempfile.NamedTemporaryFile(mode="w+b", delete=True) as savefile:
        cebra_model.save(savefile.name)
        loaded_model = cebra_sklearn_cebra.CEBRA.load(savefile.name)

    assert cebra_model.device == loaded_model.device
    assert next(cebra_model.solver_.model.parameters()).device == next(
        loaded_model.solver_.model.parameters()).device


def test_mps():
    if not cebra.helper._is_mps_availabe(torch):
        pytest.skip("MPS not available")

    X = np.random.uniform(0, 1, (10, 5))
    cebra_model = cebra_sklearn_cebra.CEBRA(model_architecture="offset1-model",
                                            max_iterations=5,
                                            device="mps")

    if torch.backends.mps.is_available() and torch.backends.mps.is_built():
        torch.backends.mps.is_available = lambda: False

        with pytest.raises(ValueError):
            cebra_model.fit(X)

        torch.backends.mps.is_available = lambda: True
        torch.backends.mps.is_built = lambda: True
        cebra_model.fit(X)
        assert hasattr(cebra_model, "n_features_")


@pytest.mark.parametrize("device", ["cuda_if_available", "cuda", "cuda:1"])
def test_cuda(device):
    if not torch.cuda.is_available():
        pytest.skip("CUDA not available")

    X = np.random.uniform(0, 1, (10, 5))
    cebra_model = cebra_sklearn_cebra.CEBRA(model_architecture="offset1-model",
                                            max_iterations=5,
                                            device=device).fit(X)

    if device == "cuda_if_available":
        cebra_model.device == "cuda"
        assert next(
            cebra_model.solver_.model.parameters()).device == torch.device(
                type='cuda', index=0)
        assert hasattr(cebra_model, "n_features_")

    elif device == "cuda":
        cebra_model.device == "cuda:0"
        assert next(
            cebra_model.solver_.model.parameters()).device == torch.device(
                type='cuda', index=0)
        assert hasattr(cebra_model, "n_features_")
    elif device == "cuda:1":
        cebra_model.device == "cuda:1"
        assert next(
            cebra_model.solver_.model.parameters()).device == torch.device(
                type='cuda', index=1)
        assert hasattr(cebra_model, "n_features_")


def test_check_device():

    device = "cuda_if_available"
    torch.cuda.is_available = lambda: True
    expected_result = "cuda"
    assert cebra_sklearn_utils.check_device(device) == expected_result

    torch.cuda.is_available = lambda: False
    cebra.helper._is_mps_availabe = lambda x: True
    expected_result = "mps"
    assert cebra_sklearn_utils.check_device(device) == expected_result

    torch.cuda.is_available = lambda: False
    cebra.helper._is_mps_availabe = lambda x: False
    expected_result = "cpu"
    assert cebra_sklearn_utils.check_device(device) == expected_result

    device = "cuda_if_available"
    torch.cuda.is_available = lambda: False
    cebra.helper._is_mps_availabe = lambda x: True
    expected_result = "mps"
    assert cebra_sklearn_utils.check_device(device) == expected_result

    device = "cuda_if_available"
    torch.cuda.is_available = lambda: False
    cebra.helper._is_mps_availabe = lambda x: False
    expected_result = "cpu"
    assert cebra_sklearn_utils.check_device(device) == expected_result

    device = "cuda:1"
    torch.cuda.device_count = lambda: 2
    expected_result = "cuda:1"
    assert cebra_sklearn_utils.check_device(device) == expected_result

    device = "cuda:2"
    torch.cuda.device_count = lambda: 2
    with pytest.raises(ValueError):
        cebra_sklearn_utils.check_device(device)

    device = "cuda:abc"
    with pytest.raises(ValueError):
        cebra_sklearn_utils.check_device(device)

    device = "cuda"
    torch.cuda.is_available = lambda: True
    expected_result = "cuda:0"
    assert cebra_sklearn_utils.check_device(device) == expected_result

    device = "cuda"
    torch.cuda.is_available = lambda: False
    with pytest.raises(ValueError):
        cebra_sklearn_utils.check_device(device)

    device = "cpu"
    expected_result = "cpu"
    assert cebra_sklearn_utils.check_device(device) == expected_result

    device = "invalid_device"
    with pytest.raises(ValueError):
        cebra_sklearn_utils.check_device(device)

    if pkg_resources.parse_version(
            torch.__version__) >= pkg_resources.parse_version("1.12"):

        device = "mps"
        torch.backends.mps.is_available = lambda: True
        torch.backends.mps.is_built = lambda: True
        expected_result = "mps"
        assert cebra_sklearn_utils.check_device(device) == expected_result

        device = "mps"
        torch.backends.mps.is_available = lambda: False
        torch.backends.mps.is_built = lambda: True
        with pytest.raises(ValueError):
            cebra_sklearn_utils.check_device(device)

        device = "mps"
        torch.backends.mps.is_available = lambda: False
        torch.backends.mps.is_built = lambda: False
        with pytest.raises(ValueError):
            cebra_sklearn_utils.check_device(device)



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Collection of helper functions that did not fit into own modules."""

import io
import pathlib
import tempfile
import urllib
import warnings
import zipfile
from functools import wraps
from typing import List, Union

import numpy as np
import numpy.typing as npt
import pkg_resources
import requests
import torch

import cebra.data


def download_file_from_url(url: str) -> str:
    """Download a fole from ``url``.

    Args:
        url: Url to fetch for the file.

    Returns:
        The path to the downloaded file.
    """
    with tempfile.NamedTemporaryFile() as tf:
        filename = tf.name + ".h5"
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    return filename


def download_file_from_zip_url(url, *, file):
    """Directly extract files without writing the archive to disk."""
    with tempfile.TemporaryDirectory() as tf:
        foldername = tf

    resp = urllib.request.urlopen(url)
    with zipfile.ZipFile(io.BytesIO(resp.read())) as zf:
        for member in zf.infolist():
            try:
                zf.extract(member, path=foldername)
            except zipfile.error:
                pass
    return pathlib.Path(foldername) / "data" / file


def _is_mps_availabe(torch):
    available = False
    if pkg_resources.parse_version(
            torch.__version__) >= pkg_resources.parse_version("1.12"):
        if torch.backends.mps.is_available():
            if torch.backends.mps.is_built():
                available = True

    return available


def _is_integer(y: Union[npt.NDArray, torch.Tensor]) -> bool:
    """Check if the values in ``y`` are :py:class:`int`.

    Args:
        y: An array, either as a :py:func:`numpy.array` or a :py:class:`torch.Tensor`.

    Returns:
        ``True`` if ``y`` contains :py:class:`int`.
    """
    return (isinstance(y, np.ndarray) and np.issubdtype(y.dtype, np.integer)
           ) or (isinstance(y, torch.Tensor) and
                 (not torch.is_floating_point(y) and not torch.is_complex(y)))


def _is_floating(y: Union[npt.NDArray, torch.Tensor]) -> bool:
    """Check if the values in ``y`` are :py:class:`float`.

    Note:
        There is no ``torch`` method to check that the ``dtype`` of a :py:class:`torch.Tensor`
        is a :py:class:`float`, consequently, we check that it is not :py:class:`int` nor
        :py:class:`complex`.

    Args:
        y: An array, either as a :py:func:`numpy.array` or a :py:class:`torch.Tensor`.

    Returns:
        ``True`` if ``y`` contains :py:class:`float`.
    """

    return (isinstance(y, np.ndarray) and
            np.issubdtype(y.dtype, np.floating)) or (isinstance(
                y, torch.Tensor) and torch.is_floating_point(y))


def _is_floating_or_integer(y: Union[npt.NDArray, torch.Tensor]) -> bool:
    """Check if the values in ``y`` are :py:class:`int` or :py:class:`float`.

    Args:
        y: An array, either as a :py:func:`numpy.array` or a :py:class:`torch.Tensor`.

    Returns:
        ``True`` if ``y`` contains :py:class:`float` or :py:class:`int`.
    """

    return _is_floating(y) or _is_integer(y)


def get_loader_options(dataset: "cebra.data.Dataset") -> List[str]:
    """Return all possible dataloaders for the given dataset.

    Notes:
        This function is deprecated and will be removed in an upcoming version of CEBRA.
        Please use :py:mod:`cebra.data.helper.get_loader_options` instead, which is an
        exact copy.
    """

    import cebra.data.helper
    warnings.warn(
        "The 'get_loader_options' function has been moved to 'cebra.data.helpers' module. "
        "Please update your imports.", DeprecationWarning)
    return cebra.data.helper.get_loader_options


def requires_package_version(module, version: str):
    """Decorator to require a minimum version of a package.

    Args:
        module: Module to be checked.
        version: The minimum required version for the module.

    Raises:
        ImportError: If the specified ``module`` version is less than
            the required ``version``.
    """

    required_version = pkg_resources.parse_version(version)

    def _requires_package_version(function):

        @wraps(function)
        def wrapper(*args, patched_version=None, **kwargs):
            if patched_version is not None:
                installed_version = pkg_resources.parse_version(
                    patched_version)  # Use the patched version if provided
            else:
                installed_version = pkg_resources.parse_version(
                    module.__version__)

            if installed_version < required_version:
                raise ImportError(
                    f"The function '{function.__name__}' requires {module.__name__} "
                    f"version {required_version} or higher, but you have {installed_version}. "
                    f"Please upgrade {module.__name__}.")
            return function(*args, **kwargs)

        return wrapper

    return _requires_package_version



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import glob
import re

import pytest

_FILENAMES = glob.glob("demo_notebooks/Demo_hippocampus.ipynb")


def _decrease_max_iterations(nb):
    for cell in nb["cells"]:
        if cell["cell_type"] == "code":
            lines = cell["source"].split("\n")
            for i in range(len(lines)):
                lines[i] = re.sub(r"max_iterations[ ]*=[ 0-9]+",
                                  "max_iterations=5", lines[i])
            cell["source"] = "\n".join(lines)
    return nb


def _change_file_path(nb):
    for cell in nb["cells"]:
        if cell["cell_type"] == "code":
            lines = cell["source"].split("\n")
            for i in range(len(lines)):
                lines[i] = re.sub("CURRENT_DIR[ ]*,", "'demo_notebooks',",
                                  lines[i])
                print(lines[i])
            cell["source"] = "\n".join(lines)
    return nb


@pytest.mark.requires_dataset
@pytest.mark.parametrize("filename", _FILENAMES)
def test_demo_notebook(filename):
    import nbformat
    from nbconvert.preprocessors import ExecutePreprocessor
    with open(filename) as f:
        nb = nbformat.read(f, as_version=4)
    nb = _decrease_max_iterations(nb)
    nb = _change_file_path(nb)
    ep = ExecutePreprocessor(timeout=600, kernel_name="python3")
    ep.preprocess(nb, {"metadata": {}})



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pathlib

import joblib as jl
import numpy as np
import torch

import cebra.data
import cebra.io
from cebra.datasets import get_datapath
from cebra.datasets import parametrize
from cebra.datasets import register

_DEFAULT_DATADIR = get_datapath()


@register("continuous-gaussian-mixture")
@parametrize(
    "continuous-gaussian-mixture-{noise}",
    noise=["poisson", "gaussian", "laplace", "uniform", "refractory_poisson"],
)
class ContinuousGaussianMixtureDataset(cebra.data.SingleSessionDataset):
    """A dataset of synthetically generated continuous labels and the corresponding 2D latents
    and 100D noisy observations.

    Args:
        noise: The applied noise distribution applied.
    """

    def __init__(self, noise: str = "poisson"):
        super().__init__()
        self.noise = noise
        data = jl.load(
            pathlib.Path(_DEFAULT_DATADIR) / "synthetic" /
            f"continuous_label_{self.noise}.jl")
        self.latent = data["z"]
        self.index = torch.from_numpy(data["u"]).float()
        self.neural = torch.from_numpy(data["x"]).float()

    @property
    def input_dimension(self):
        return self.neural.size(1)

    @property
    def continuous_index(self):
        return self.index

    def __getitem__(self, index):
        index = self.expand_index(index)
        return self.neural[index].transpose(2, 1)

    def split(self, split, train_ratio=0.8, valid_ratio=1):
        tot_len = len(self.neural)
        train_idx = np.arange(tot_len)[:int(tot_len * train_ratio)]
        valid_idx = np.arange(tot_len)[int(tot_len * train_ratio):]

        if split == "train":
            self.neural = self.neural[train_idx]
            self.index = self.index[train_idx]

        elif split == "valid":
            self.neural = self.neural[valid_idx]
            self.index = self.index[valid_idx]

        elif split == "all":
            pass

    def __len__(self):
        return len(self.neural)

    def __repr__(self):
        return f"ContinuousGaussianMixtureDataset(noise: {self.noise}, shape: {self.neural.shape})"



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Index operations for conditional sampling.

Indexing operations---in contrast to data distributions---exhibit deterministic behavior
by returning an element closest in the dataset to a given query sample. This module contains
helper functions for mixed and continuously indexed datasets (i.e., containing discrete and/or
continuous data).

Discrete data has to come in the format of a single label for each datapoint. Multidimensional
discrete labels should be converted accordingly.
"""

import torch

import cebra.data
import cebra.distributions.base as cebra_distributions
import cebra.io

_INF = float("inf")


def _is_float_tensor(tensor):
    return isinstance(tensor, torch.Tensor) and torch.is_floating_point(tensor)


def _check_is_float_tensor(sender, tensor):
    if not _is_float_tensor(tensor):
        raise ValueError(
            f"{sender} requires a torch.Tensor of floating point type "
            f"(either on cpu or gpu), but got {type(tensor)} with dtype={tensor.dtype}."
        )
    return isinstance(tensor, torch.Tensor) and torch.is_floating_point(tensor)


class DistanceMatrix(cebra.io.HasDevice):
    """Compute shortest distances between dataset samples.

    Args:
        samples: The continuous values that will be used to index
            the dataset and specify the conditional distribution.

    Note:
        This implementation is not particularly efficient on very
        large datasets. For these cases, packages like FAISS offer
        more optimized retrieval functions.

        As a rule of thumb, using this class is suitable for
        datasets for which the dataset can be hosted on GPU memory.
    """

    def __init__(self, samples: torch.Tensor):
        _check_is_float_tensor(self, samples)
        self.index = samples
        self.xTx = self.index.square().sum(1, keepdim=True)

    def __call__(self, query, mask=None):
        """Compute the pairwise distances between index and query.

        Args:
            query: (n, d)
                The query matrix
            mask: (N,)
                A binary mask with same length as the index

        Returns: (M,n)
            Pairwise distances between the m samples with True binary
            mask (default: N) and the given query samples.
        """
        # TODO(stes): slow
        query = query.to(self.device)

        qTq = query.square().sum(1, keepdim=True).T
        if mask is None:
            xTq = torch.einsum("ni,mi->nm", self.index, query)
            xTx = self.xTx
        else:
            xTq = torch.einsum("ni,mi->nm", self.index[mask], query)
            xTx = self.xTx[mask]
        return xTx + qTq - 2 * xTq


class OffsetDistanceMatrix(DistanceMatrix):
    """Compute shortest distances, ignoring samples close to the boundary.

    Compared to the standard ``DistanceMatrix``, this class should be used
    for datasets and learning setups where multiple timesteps are fed into
    the network at once --- the samples close to the time-series boundary
    should be ignored in the sampling process in these cases.

    Args:
        samples: The continuous values that will be used to index
            the dataset and specify the conditional distribution.
        offset: The number of timesteps to ignored at each size of the
            dataset

    TODO:
        * switch offset to `cebra.data.Offset`
    """

    def __init__(self, samples, offset: int = 1):
        super().__init__(samples)
        self.inf = torch.tensor(_INF)
        self.offset = cebra.data.Offset(offset)
        if len(self.offset) < 1:
            raise ValueError(
                f"Choose an offset of at least 1, otherwise use {type(super())}."
            )
        self.offset.mask_array(self.xTx, self.inf)


class ContinuousIndex(cebra_distributions.Index, cebra.io.HasDevice):
    """Naive nearest neighbor search implementation.

    index: tensor(N, d)
        the values used for kNN search
    offset: int or (int,int)
        the time offset in each direction
    """

    def __init__(self, index):
        super().__init__()
        _check_is_float_tensor(self, index)
        self.dist_matrix = DistanceMatrix(index)

    def search(self, query):
        """Return index location closest to query."""
        distance = self.dist_matrix(query)
        return torch.argmin(distance, dim=0)
        # TODO(stes) handle offsets
        # + self.dist_matrix.offset.left


class ConditionalIndex(cebra_distributions.Index):
    """Index a dataset based on both continuous and discrete information.

    In contrast to the standard :py:class:`.base.Index` class, the :py:class:`ConditionalIndex`
    accept both discrete and continuous indexing information.

    This index considers the discrete indexing information first to
    identify possible positive pairs. Then among these candidate samples,
    behaves like an :py:class:`.base.Index` and returns the samples closest in terms of
    the information in the continuous index.

    Args:
        discrete: The discrete indexing information, which should be
            limited to a 1d feature space. If higher dimensional discrete
            vectors are used, they should be first re-formatted to fit
            this structure.
        continuous: The continuous indexing information, which can be a
            vector of arbitrary dimension and will be used to define the
            distance between the samples that share the same discrete
            index.
    """

    def __init__(self, discrete, continuous):
        _check_is_float_tensor(self, continuous)
        if discrete is None:
            raise ValueError(
                "The specified discrete index was set to None. "
                "If this was intended, use Index instead of ConditionalIndex "
                "which does not require to specify discrete indexing "
                "information.")
        if len(discrete) != len(continuous):
            raise ValueError(
                f"Discrete ({len(discrete)} samples) and continuous "
                f"({len(continuous)} samples) need to match in their number "
                "of samples.")
        if len(discrete.shape) > 1:
            raise ValueError(
                "Discrete indexing information needs to be limited to a 1d "
                "array/tensor. Multi-dimensional discrete indices should be "
                "reformatted first.")
            # TODO(stes): Once a helper function exists, the error message should
            #            mention it.

        self.discrete = discrete
        self.continuous = continuous

        self.distance_matrix = DistanceMatrix(self.continuous)

        self.mask_x = {
            int(v): (self.discrete == v) for v in torch.unique(discrete)
        }
        self.mask_idx = {
            int(v): torch.nonzero(mask_x).squeeze()
            for v, mask_x in self.mask_x.items()
        }

    def search(self, continuous, discrete=None):
        """Search closest sample based on continuous and discrete indexing
        information.

        Args:
            continuous:
                Samples from the continuous index
            discrete:
                Optionally matching samples from the discrete index,
                used to pre-select matching indices.
        """
        if continuous.shape[1] != self.continuous.shape[1]:
            raise ValueError(f"Shape of continuous index does not match along "
                             f"the feature dimension. "
                             f"Expected {self.continuous.shape[1]}d, but got "
                             f"{continuous.shape[1]}d.")
        if discrete is None:
            return self.search_naive(continuous, discrete=None)

        # TODO(stes) select based on expected speed advantage
        return self.search_naive(continuous, discrete)

    def __getitem__(self, value):
        # TODO(stes): this function might not be used; consider removing
        #            for removing, tests should pass while this function
        #            returns a deprecation exception; it is unclear why
        #            this adds an overall value to the API.
        continuous, discrete = value
        return self.search(continuous, discrete)

    def search_naive(self, continuous, discrete):
        """Brute force search
        Fast especially for small indices

        Args:
            continuous:
                TODO
            discrete:
                TODO
        """
        # TODO(stes): slow
        continuous = continuous.to(self.device)

        distance = self.distance_matrix(continuous)
        if discrete is not None:
            # TODO(stes): slow
            discrete = discrete.to(self.device)
            mask = torch.eq(self.discrete[:, None], discrete[None, :])
            distance[~mask] = _INF  # TODO: inefficient
        return torch.argmin(distance, dim=0)

    def search_iterative(self, continuous, discrete):
        """Iterative search
        Gets faster especially for >1e6 samples in the index.

        Args:
            continuous:
                TODO
            discrete:
                TODO
        """
        # TODO(stes): slow
        discrete = discrete.to(self.device)
        continuous = continuous.to(self.device)

        ret = torch.zeros_like(discrete)
        for v in torch.unique(discrete):
            mask_x = self.mask_x[int(v)]
            mask_q = discrete == v
            diff = self.distance_matrix(continuous[mask_q], mask=mask_x)
            ret[mask_q] = self.mask_idx[int(v)][torch.argmin(diff, dim=0)]
        return ret


class MultiSessionIndex(cebra_distributions.Index):
    """Index multiple sessions.

    Args:
        indices: Indices for the different sessions. Indices of multi-session
            datasets should have matching feature dimension.
    """

    def __init__(self, *indices):
        self.indices = indices

    def search(self, query):
        """Return closest element in each of the datasets.

        Args:
            query: The query which is applied to each index of
                the dataset.

        Returns:
            A list of indices from each session.
        """
        return [index.search(query) for index in self.indices]



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import functools
from typing import Literal, Optional

import numpy as np
import pytest
import torch

import cebra.datasets as cebra_datasets
import cebra.distributions as cebra_distr
import cebra.distributions.base as cebra_distr_base


def assert_is_tensor(T, device=None):
    assert isinstance(T, torch.Tensor)
    if device is not None:
        assert T.device == device


def prepare(N=1000, n=128, d=5, probs=[0.3, 0.1, 0.6], device="cpu"):
    discrete = torch.from_numpy(np.random.choice([0, 1, 2], p=probs,
                                                 size=(N,))).to(device)
    continuous = torch.randn(N, d).to(device)

    rand = torch.from_numpy(np.random.randint(0, N, (n,))).to(device)
    _ = discrete[rand].to(device)
    query = continuous[rand] + 0.1 * torch.randn(n, d).to(device)
    query = query.to(device)

    return discrete, continuous


def test_init_discrete():
    discrete, continuous = prepare()

    # Sampling operations on discrete data distributions
    cebra_distr.discrete.Discrete(discrete)
    cebra_distr.discrete.DiscreteUniform(discrete)
    cebra_distr.discrete.DiscreteEmpirical(discrete)


def test_distance_matrix():
    _, continuous = prepare()
    matrix = cebra_distr.index.DistanceMatrix(continuous)

    query = continuous[:10]
    dist = matrix(query)
    idx = torch.argmin(dist, dim=0)
    assert dist.shape == (len(continuous), 10)
    assert idx.shape == (10,)

    assert torch.allclose(continuous[idx], continuous[:10])
    assert torch.eq(idx, torch.arange(10)).all()


def test_init_index():
    discrete, continuous = prepare()

    # Indexing operations on continuous data
    idx = cebra_distr.index.ContinuousIndex(continuous)
    cidx = cebra_distr.index.ConditionalIndex(discrete, continuous)

    a = idx.search(continuous[:10])
    assert a.shape == (10,)
    assert torch.allclose(continuous[a], continuous[:10])
    assert torch.eq(a, torch.arange(10)).all()

    cidx = cebra_distr.index.ConditionalIndex(discrete, continuous)
    b = cidx.search(continuous[:10], discrete[:10])
    assert b.shape == (10,)
    assert torch.eq(b, torch.arange(10)).all()
    assert torch.allclose(continuous[b], continuous[:10])
    assert torch.eq(b, torch.arange(10)).all()


class _TestMixedBase:

    @functools.cached_property
    def ref_idx(self):
        return self.distribution.sample_prior(self.num_samples)

    def setup_method(self):
        self.num_samples = 15
        self.discrete, self.continuous = prepare()

    def test_prior(self):
        self.setup_method()
        assert_is_tensor(self.discrete)
        assert_is_tensor(self.continuous)
        assert_is_tensor(self.ref_idx)

    def test_conditional(self):
        self.setup_method()
        # The conditional distribution p(· | disc, cont) should yield
        # samples where the label exactly matches the reference sample.
        samples_both = self.distribution.sample_conditional(
            self.discrete[self.ref_idx], self.continuous[self.ref_idx])
        assert_is_tensor(samples_both)

        # the discrete labels should exactly match for all samples
        assert torch.eq(self.discrete[samples_both],
                        self.discrete[self.ref_idx]).all()

        # the returned indices should differ from the reference indices
        assert not torch.eq(samples_both, self.ref_idx).any()


class TestMixed(_TestMixedBase):

    def setup_method(self):
        super().setup_method()
        self.distribution = cebra_distr.mixed.Mixed(self.discrete,
                                                    self.continuous)

    def test_conditional(self):
        pytest.skip("Skipping")

    def test_conditional_continuous(self):
        pytest.skip("Skipping")

        self.setup_method()
        # Sampling only based on the continuous samples, p(· | cont) should reproduce
        # the empirical distribution of discrete values.
        samples_cont = self.distribution.sample_conditional_continuous(
            self.continuous[self.ref_idx])
        assert_is_tensor(samples_cont)

        # the returned indices should differ from the reference indices
        assert not torch.eq(samples_cont, self.ref_idx).any()

        # the discrete labels are not all same, since we marginalize across the
        # discrete values
        assert not torch.eq(self.discrete[samples_cont],
                            self.discrete[self.ref_idx]).all()

    def test_conditional_discrete(self):
        self.setup_method()
        samples_disc = self.distribution.sample_conditional_discrete(
            self.discrete[self.ref_idx])
        assert_is_tensor(samples_disc)

        # a few samples can be the same, but not all of them (extremely unlikely)
        assert not all(samples_disc == self.ref_idx)

        # the discrete labels should exactly match for all samples
        assert torch.eq(self.discrete[samples_disc],
                        self.discrete[self.ref_idx]).all()


def test_mixed():
    discrete, continuous = prepare()
    distribution = cebra_distr.mixed.MixedTimeDeltaDistribution(
        discrete, continuous)

    reference_idx = distribution.sample_prior(10)
    _ = distribution.sample_conditional(reference_idx)

    # The conditional distribution p(· | disc, cont) should yield
    # samples where the label exactly matches the reference sample.
    samples_both = distribution.sample_conditional(reference_idx)
    assert_is_tensor(samples_both)

    # the discrete labels should exactly match for all samples
    assert torch.eq(discrete[samples_both], discrete[reference_idx]).all()

    # the returned indices should differ from the reference indices
    assert not torch.eq(samples_both, reference_idx).all()


def test_continuous(benchmark):
    discrete, continuous = prepare()

    def _test_distribution(dist):
        distribution = dist(continuous)
        reference_idx = distribution.sample_prior(10)
        _ = distribution.sample_conditional(reference_idx)
        return distribution

    distribution = _test_distribution(
        cebra_distr.continuous.TimedeltaDistribution)

    def _conditional():
        reference_idx = distribution.sample_prior(1000)
        distribution.sample_conditional(reference_idx)

    benchmark(_conditional)

    distribution = _test_distribution(cebra_distr.continuous.TimeContrastive)


def test_discrete_uniform_discrete():
    """Functional test for empirical/uniform modes"""
    np.random.seed(0)
    N = 10000
    probs = [0.3, 0.1, 0.6]
    samples = np.random.choice([0, 1, 2], p=probs, size=(N,))
    dist = cebra_distr.Discrete(samples)
    resample_uni = samples[dist.sample_uniform(N)]
    resample_emp = samples[dist.sample_empirical(N)]

    assert samples.shape == (N,)
    assert resample_uni.shape == (N,)
    assert resample_emp.shape == (N,)

    assert np.allclose(
        np.bincount(samples) / N, np.array([0.3055, 0.0974, 0.5971]))
    assert np.allclose(
        np.bincount(resample_uni) / N, np.array([0.3424, 0.3278, 0.3298]))
    assert np.allclose(
        np.bincount(resample_emp) / N, np.array([0.2969, 0.098, 0.6051]))


@pytest.mark.parametrize("time_offset", [1, 5, 10])
def test_single_session_time_contrastive(time_offset):
    """Single session time-contrastive learning.

    The test checks if sampled pairs have the correct time offset.
    TODO: check kfold datasets.
    """

    index = torch.arange(100).unsqueeze(1)
    distribution = cebra_distr.TimeContrastive(index, time_offset=time_offset)

    num_samples = 5
    sample = distribution.sample_prior(num_samples)
    assert sample.shape == (num_samples,)

    positive = distribution.sample_conditional(sample)
    assert torch.eq(positive - sample, time_offset).all()
    assert positive.shape == (num_samples,)


@pytest.mark.parametrize("index_difference", [1, 5, 10])
def test_single_session_time_delta(index_difference):
    index = torch.cumsum(torch.ones(100) * index_difference,
                         dim=0).unsqueeze(1).float()
    distribution = cebra_distr.TimedeltaDistribution(index)

    num_samples = 5
    sample = distribution.sample_prior(num_samples)
    assert sample.shape == (num_samples,)

    positive = distribution.sample_conditional(sample)
    assert positive.shape == (num_samples,)

    assert not torch.eq(positive, sample).all(
    ), "No samples have time delta of 1, hence all indices should be different"

    sample = index[sample]
    positive = index[positive]

    assert (abs(positive - sample) <= index_difference).all()


def test_single_session_discrete():
    pass


def test_multi_session_time_delta():
    pass


@pytest.mark.parametrize("time_offset", [1, 5, 10])
def test_multi_session_time_contrastive(time_offset):
    dataset = cebra_datasets.init("demo-continuous-multisession")
    sampler = cebra_distr.MultisessionSampler(dataset, time_offset=time_offset)

    num_samples = 5
    sample = sampler.sample_prior(num_samples)
    assert sample.shape == (dataset.num_sessions, num_samples)

    positive, idx, rev_idx = sampler.sample_conditional(sample)
    assert positive.shape == (dataset.num_sessions, num_samples)
    assert idx.shape == (dataset.num_sessions * num_samples,)
    assert rev_idx.shape == (dataset.num_sessions * num_samples,)
    # NOTE(celia): test the private function ``_inverse_idx()``, with idx arrays flat
    assert (idx.flatten()[rev_idx.flatten()].all() == np.arange(
        len(rev_idx.flatten())).all())


def test_multi_session_discrete():
    dataset = cebra_datasets.init("demo-discrete-multisession")
    sampler = cebra_distr.DiscreteMultisessionSampler(dataset)

    num_samples = 5
    sample = sampler.sample_prior(num_samples)
    assert sample.shape == (dataset.num_sessions, num_samples)

    positive, idx, rev_idx = sampler.sample_conditional(sample)
    assert positive.shape == (dataset.num_sessions, num_samples)
    assert idx.shape == (dataset.num_sessions * num_samples,)
    assert rev_idx.shape == (dataset.num_sessions * num_samples,)
    # NOTE(celia): test the private function ``_inverse_idx()``, with idx arrays flat
    assert (idx.flatten()[rev_idx.flatten()].all() == np.arange(
        len(rev_idx.flatten())).all())

    # Check positive samples' labels match reference samples' labels
    sample_labels = sampler.all_data[(sample +
                                      sampler.lengths[:, None]).flatten()]
    sample_labels = sample_labels[idx.reshape(sample.shape[:2])].flatten()
    positive_labels = sampler.all_data[(positive +
                                        sampler.lengths[:, None]).flatten()]
    assert (sample_labels == positive_labels).all()


class OldDeltaDistribution(cebra_distr_base.JointDistribution,
                           cebra_distr_base.HasGenerator):
    """
    Old version of the Delta Distribution where it only works for 1d
    behavior variable.

    """

    def __init__(self,
                 continuous: torch.Tensor,
                 delta: float = 0.1,
                 device: Literal["cpu", "cuda"] = "cpu",
                 seed: Optional[int] = 1812):
        cebra_distr_base.HasGenerator.__init__(self, device=device, seed=seed)
        torch.manual_seed(seed)
        self.data = continuous
        self.std = delta
        self.index = cebra_distr.ContinuousIndex(self.data)
        self.prior = cebra_distr.Prior(self.data, device=device, seed=seed)

    def sample_prior(self, num_samples: int) -> torch.Tensor:
        """See :py:meth:`.Prior.sample_prior`."""
        return self.prior.sample_prior(num_samples)

    def sample_conditional(self, reference_idx: torch.Tensor) -> torch.Tensor:
        """Return indices from the conditional distribution."""

        if reference_idx.dim() != 1:
            raise ValueError(
                f"Reference indices have wrong shape: {reference_idx.shape}. "
                "Pass a 1D array of indices of reference samples.")

        # TODO(stes): Set seed
        query = torch.distributions.Normal(
            self.data[reference_idx].squeeze(),
            torch.ones_like(reference_idx, device=self.device) * self.std,
        ).sample()

        return self.index.search(query.unsqueeze(-1))


def test_old_vs_new_delta_normal_with_1Dindex():
    _, continuous = prepare()
    assert continuous.dim() == 2
    num_samples = len(continuous)
    reference_idx = torch.randint(0, num_samples, (num_samples,))

    new_distribution = cebra_distr.DeltaNormalDistribution(
        continuous=continuous[:, 0].unsqueeze(-1), delta=0.1)

    old_distribution = OldDeltaDistribution(
        continuous=continuous[:, 0].unsqueeze(-1), delta=0.1)

    torch.manual_seed(1812)
    old_positives = old_distribution.sample_conditional(reference_idx)
    torch.manual_seed(1812)
    new_positives = new_distribution.sample_conditional(reference_idx)

    assert not torch.equal(old_positives, reference_idx)
    assert not torch.equal(new_positives, reference_idx)
    assert torch.equal(old_positives, new_positives)


@pytest.mark.parametrize("delta,numerical_check", [(0.01, True), (0.025, True),
                                                   (1., False), (5., False)])
def test_new_delta_normal_with_multidimensional_index(delta, numerical_check):
    continuous = torch.rand(100_000, 3).to("cpu")
    num_samples = 1000
    delta_normal_multidim = cebra_distr.DeltaNormalDistribution(
        delta=delta, continuous=continuous)
    reference_idx = delta_normal_multidim.sample_prior(num_samples)
    positive_idx = delta_normal_multidim.sample_conditional(reference_idx)

    assert positive_idx.dim() == 1
    assert len(positive_idx) == num_samples
    assert not torch.equal(positive_idx, reference_idx)

    if numerical_check:
        reference_samples = continuous[reference_idx]
        positive_samples = continuous[positive_idx]
        diff = positive_samples - reference_samples
        #TODO(stes): Improve test, use lower error margin here
        assert torch.isclose(diff.std(), torch.tensor(delta), rtol=0.1)
    else:
        #TODO(stes): Add a warning message to the delta distribution.
        pytest.skip(
            "multivariate delta distribution can not accurately sample with the "
            "given parameters. TODO: Add a warning message for these cases.")



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Base classes for datasets and loaders."""

import abc

import literate_dataclasses as dataclasses
import torch

import cebra.data.assets as cebra_data_assets
import cebra.distributions
import cebra.io
from cebra.data.datatypes import Batch
from cebra.data.datatypes import BatchIndex
from cebra.data.datatypes import Offset

__all__ = ["Dataset", "Loader"]


class Dataset(abc.ABC, cebra.io.HasDevice):
    """Abstract base class for implementing a dataset.

    The class attributes provide information about the shape of the data when
    indexing this dataset.

    Attributes:
        input_dimension: The input dimension of the signal in this dataset.
            Models applied on this this dataset should match this dimensionality.
        offset: The offset determines the shape of the data obtained with the
            ``__getitem__`` and :py:meth:`expand_index` methods.
    """

    def __init__(self,
                 device="cpu",
                 download=False,
                 data_url=None,
                 data_checksum=None,
                 location=None,
                 file_name=None):

        self.offset: Offset = cebra.data.Offset(0, 1)
        super().__init__(device)

        self.download = download
        self.data_url = data_url
        self.data_checksum = data_checksum
        self.location = location
        self.file_name = file_name

        if self.download:
            if self.data_url is None:
                raise ValueError(
                    "Missing data URL. Please provide the URL to download the data."
                )

            if self.data_checksum is None:
                raise ValueError(
                    "Missing data checksum. Please provide the checksum to verify the data integrity."
                )

            cebra_data_assets.download_file_with_progress_bar(
                url=self.data_url,
                expected_checksum=self.data_checksum,
                location=self.location,
                file_name=self.file_name)

    @property
    @abc.abstractmethod
    def input_dimension(self) -> int:
        raise NotImplementedError

    @property
    def continuous_index(self) -> torch.Tensor:
        """The continuous index, if available.

        The continuous index along with a similarity metric is used for drawing
        positive and/or negative samples.

        Returns:
            Tensor of shape ``(N,d)``, representing the
            index for all ``N`` samples in the dataset.
        """
        return None

    @property
    def discrete_index(self) -> torch.Tensor:
        """The discrete index, if available.

        The discrete index can be used for making an embedding invariant to
        a variable for to restrict positive samples to share the same index variable.
        To implement more complicated indexing operations (such as modeling similiarities
        between indices), it is better to transform a discrete into a continuous index.

        Returns:
            Tensor of shape ``(N,)``, representing the index
            for all ``N`` samples in the dataset.
        """
        return None

    def expand_index(self, index: torch.Tensor) -> torch.Tensor:
        """

        Args:
            index: A one-dimensional tensor of type long containing indices
                to select from the dataset.

        Returns:
            An expanded index of shape ``(len(index), len(self.offset))`` where
            the elements will be
            ``expanded_index[i,j] = index[i] + j - self.offset.left`` for all ``j``
            in ``range(0, len(self.offset))``.

        Note:
            Requires the :py:attr:`offset` to be set.
        """

        # TODO(stes) potential room for speed improvements by pre-allocating these tensors/
        # using non_blocking copy operation.
        offset = torch.arange(-self.offset.left,
                              self.offset.right,
                              device=index.device)

        index = torch.clamp(index, self.offset.left,
                            len(self) - self.offset.right)

        return index[:, None] + offset[None, :]

    def expand_index_in_trial(self, index, trial_ids, trial_borders):
        """When the neural/behavior is in discrete trial, e.g) Monkey Reaching Dataset
        the slice should be defined within the trial.
        trial_ids is in size of a length of self.index and indicate the trial id of the index belong to.
        trial_borders is in size of a length of self.idnex and indicate the border of each trial.

        Todo:
            - rewrite
        """

        # TODO(stes) potential room for speed improvements by pre-allocating these tensors/
        # using non_blocking copy operation.
        offset = torch.arange(-self.offset.left,
                              self.offset.right,
                              device=index.device)
        index = torch.tensor(
            [
                torch.clamp(
                    i,
                    trial_borders[trial_ids[i]] + self.offset.left,
                    trial_borders[trial_ids[i] + 1] - self.offset.right,
                ) for i in index
            ],
            device=self.device,
        )
        return index[:, None] + offset[None, :]

    @abc.abstractmethod
    def __getitem__(self, index: torch.Tensor) -> torch.Tensor:
        """Return samples at the given time indices.

        Args:
            index: An indexing tensor of type :py:attr:`torch.long`.

        Returns:
            Samples from the dataset matching the shape
            ``(len(index), self.input_dimension, len(self.offset))``
        """

        raise NotImplementedError

    @abc.abstractmethod
    def load_batch(self, index: BatchIndex) -> Batch:
        """Return the data at the specified index location.

        TODO: adapt signature to support Batches and List[Batch]
        """
        raise NotImplementedError()

    def configure_for(self, model: "cebra.models.Model"):
        """Configure the dataset offset for the provided model.

        Call this function before indexing the dataset. This sets the
        :py:attr:`offset` attribute of the dataset.

        Args:
            model: The model to configure the dataset for.
        """
        self.offset = model.get_offset()


@dataclasses.dataclass
class Loader(abc.ABC, cebra.io.HasDevice):
    """Base dataloader class.

    Args:
        See dataclass fields.

    Yields:
        Batches of the specified size from the given dataset object.

    Note:
        The ``__iter__`` method is non-deterministic, unless explicit seeding is implemented
        in derived classes. It is recommended to avoid global seeding in numpy
        and torch, and instead locally instantiate a ``Generator`` object for
        drawing samples.
    """

    dataset: Dataset = dataclasses.field(
        default=None,
        doc="""A dataset instance specifying a ``__getitem__`` function.""",
    )

    num_steps: int = dataclasses.field(
        default=None,
        doc=
        """The total number of batches when iterating over the dataloader.""",
    )

    batch_size: int = dataclasses.field(default=None,
                                        doc="""The total batch size.""")

    def __post_init__(self):
        if self.num_steps is None or self.num_steps <= 0:
            raise ValueError(
                f"num_steps cannot be less than or equal to zero or None. Got {self.num_steps}"
            )
        if self.batch_size is not None and self.batch_size <= 0:
            raise ValueError(
                f"Batch size has to be None, or a non-negative value. Got {self.batch_size}."
            )

    def __len__(self):
        """The number of batches returned when calling as an iterator."""
        return self.num_steps

    def __iter__(self) -> Batch:
        for _ in range(len(self)):
            index = self.get_indices(num_samples=self.batch_size)
            yield self.dataset.load_batch(index)

    @abc.abstractmethod
    def get_indices(self, num_samples: int):
        """Sample and return the specified number of indices.

        The elements of the returned `BatchIndex` will be used to index the
        `dataset` of this data loader.

        Args:
            num_samples: The size of each of the reference, positive and
                negative samples.

        Returns:
            batch indices for the reference, positive and negative sample.
        """
        raise NotImplementedError()



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import hashlib
import re
import warnings
from pathlib import Path
from typing import Optional

import requests
import tqdm

_MAX_RETRY_COUNT = 2


def download_file_with_progress_bar(url: str,
                                    expected_checksum: str,
                                    location: str,
                                    file_name: str,
                                    retry_count: int = 0) -> Optional[str]:
    """Download a file from the given URL.

    During download, progress is reported using a progress bar. The downloaded
    file's checksum is compared to the provided ``expected_checksum``.

    Args:
        url: The URL to download the file from.
        expected_checksum: The expected checksum value of the downloaded file.
        location: The directory where the file will be saved.
        file_name: The name of the file.
        retry_count: The number of retry attempts (default: 0).

    Returns:
        The path of the downloaded file if the download is successful, None otherwise.

    Raises:
        RuntimeError: If the maximum ``retry count`` is exceeded.
    """

    # Check if the file already exists in the location
    location_path = Path(location)
    file_path = location_path / file_name

    if file_path.exists():
        existing_checksum = calculate_checksum(file_path)
        if existing_checksum == expected_checksum:
            return file_path

    if retry_count >= _MAX_RETRY_COUNT:
        raise RuntimeError(
            f"Exceeded maximum retry count ({_MAX_RETRY_COUNT}). "
            f"Unable to download the file from {url}")

    response = requests.get(url, stream=True)

    # Check if the request was successful
    if response.status_code != 200:
        raise requests.HTTPError(
            f"Error occurred while downloading the file. Response code: {response.status_code}"
        )

    # Check if the response headers contain the 'Content-Disposition' header
    if 'Content-Disposition' not in response.headers:
        raise ValueError(
            "Unable to determine the filename. 'Content-Disposition' header not found."
        )

    # Extract the filename from the 'Content-Disposition' header
    filename_match = re.search(r'filename="(.+)"',
                               response.headers.get("Content-Disposition"))
    if not filename_match:
        raise ValueError(
            "Unable to determine the filename from the 'Content-Disposition' header."
        )

    # Create the directory and any necessary parent directories
    location_path.mkdir(exist_ok=True)

    filename = filename_match.group(1)
    file_path = location_path / filename

    total_size = int(response.headers.get("Content-Length", 0))
    checksum = hashlib.md5()  # create checksum

    with open(file_path, "wb") as file:
        with tqdm.tqdm(total=total_size, unit="B",
                       unit_scale=True) as progress_bar:
            for data in response.iter_content(chunk_size=1024):
                file.write(data)
                checksum.update(
                    data)  # Update the checksum with the downloaded data
                progress_bar.update(len(data))

    downloaded_checksum = checksum.hexdigest()  # Get the checksum value
    if downloaded_checksum != expected_checksum:
        warnings.warn(f"Checksum verification failed. Deleting '{file_path}'.")
        file_path.unlink()
        warnings.warn("File deleted. Retrying download...")

        # Retry download using a for loop
        for _ in range(retry_count + 1, _MAX_RETRY_COUNT + 1):
            return download_file_with_progress_bar(url, expected_checksum,
                                                   location, file_name,
                                                   retry_count + 1)
    else:
        print(f"Download complete. Dataset saved in '{file_path}'")
        return url


def calculate_checksum(file_path: str) -> str:
    """Calculate the MD5 checksum of a file.

    Args:
        file_path: The path to the file.

    Returns:
        The MD5 checksum of the file.
    """
    checksum = hashlib.md5()
    with open(file_path, "rb") as file:
        for chunk in iter(lambda: file.read(4096), b""):
            checksum.update(chunk)
    return checksum.hexdigest()



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helper classes and functions for I/O functionality."""

import joblib
import numpy as np
import sklearn.decomposition
import torch
from torch import nn

if torch.cuda.is_available():
    _device = "cuda"
else:
    _device = "cpu"


def device() -> str:
    """The preferred compute device.

    Returns:
        ``cuda``, if available, otherwise ``cpu``.
    """
    return _device


class HasDevice:
    """Base class for classes that use CPU/CUDA processing via PyTorch.

    If implementing this class, any derived instanced will track any attribute
    of types ``torch.Tensor`` (this includes ``torch.nn.Parameter``),
    ``torch.nn.Module`` as well as any class subclassing :py:class:`HasDevice`
    itself.

    When calling :py:meth:`to`, all of these attributes will themselves be moved
    to the specified ``device``.

    Any instance of this class will need to be initialized. This can happen explicitly
    through the constructor (by specifying a device during initialization) or by assigning
    the first element, which will assign the device of the first element to the whole
    instance.

    Every following assignment will yield in tensors, parameters, modules and other
    instances being moved to the device of the instance.

    Args:
        device: The device name, typically ``cpu`` or ``cuda``, or combined
            with a device idea, e.g. ``cuda:0``.

    Note:
        Do not subclass this class when some dependency to a compute device
        is already implemented, e.g. via a pytorch ``torch.nn.Module``.
    """

    def __init__(self, device: str = None):
        if device is not None:
            self._init(device)

    def _init(self, device):
        self._tensors = set()
        self._children = set()
        self._modules = set()
        self._set_device(device)

    def _set_device(self, device):
        if device is None:
            return
        if not isinstance(device, str):
            device = device.type
        if device not in ("cpu", "cuda", "mps"):
            if device.startswith("cuda"):
                _, id_ = device.split(":")
                if int(id_) >= torch.cuda.device_count():
                    raise ValueError(device)
            else:
                raise ValueError(device)
        self._device = device

    @property
    def _initialized(self) -> bool:
        """

        ``True`` if the instance was either initialized upon creation via
        passing the ``device`` option to the constructor, or later by setting
        any attribute of the instance. The device of the instance will then
        be set to the first element's device.
        """
        return hasattr(self, "_modules")

    def _assert_initialized(self, device=None):
        if not self._initialized:
            self._init(device=device)
        assert self._initialized
        # if dataclasses.is_dataclass(self):
        # else:
        #    raise RuntimeError(f"{self} not initialized. Did you call super().__init__?")

    @property
    def device(self) -> str:
        """Returns the name of the current compute device."""
        self._assert_initialized()
        return self._device

    def to(self, device: str) -> "HasDevice":
        """Moves the instance to the specified device.

        Args:
            device: The device (`cpu` or `cuda`) to move this instance to.

        Returns:
            the instance itself.
        """
        self._assert_initialized()
        self._set_device(device)
        if device is None:
            return self
        for tensor in self._tensors:
            setattr(self, tensor, getattr(self, tensor))
        for child in self._children:
            getattr(self, child).to(device)
        for module in self._modules:
            getattr(self, module).to(device)
        return self

    def __setattr__(self, property, value):
        if isinstance(value, torch.Tensor):
            self._assert_initialized(device=value.device.type)
            self._tensors.add(property)
            value = value.to(self.device)
            assert value is not None
        elif isinstance(value, HasDevice):
            self._assert_initialized(device=value.device)
            self._children.add(property)
            value = value.to(self.device)
            assert value is not None
        elif isinstance(value, nn.Module):
            param = None
            for param in value.parameters():
                break  # Only get the first parameter
            if param is not None:
                self._assert_initialized(device=param.device.type)
                self._modules.add(property)
                value.to(self.device)
            assert value is not None
        if value is None:
            self._assert_initialized()
            for container in (self._tensors, self._children, self._modules):
                if property in container:
                    print(f"Remove {property}")
                    container.remove(property)
                    break
        if property == "device":
            raise ValueError(
                f"Cannot set the device directly. Instead, call {self}.to(device)."
            )
        super().__setattr__(property, value)


def reduce(data, *, ratio=None, num_components=None):
    """Map the specified data to its principal components

    Specify either an explained variance ``ratio`` between 0 and 1,
    or a number of principle components to use.

    Args:
        ratio: The ratio (needs to be between 0 and 1) of explained variance
            required by the returned number of components. Note that the
            dimension of the output will vary based on the provided input
            data.
        num_components: The number of principal components to return

    Returns:
        An ``(N, d)`` array, where the dimension ``d`` is either limited by
        the specified number of components, or is chosen to explain the specified
        variance in the data.

    """
    if (ratio is None) and (num_components is None):
        raise ValueError(
            "Specify either a threshold on the explained variance, or a maximum"
            "number of principle components")
    pca = sklearn.decomposition.PCA(num_components)
    data = data.reshape(len(data), -1)
    pca.fit(data)
    if ratio is None:
        i = None
    else:
        if ratio <= 0 or ratio > 1:
            raise ValueError(f"Ratio must be in (0, 1], but got {ratio}.")
        i = np.where(np.cumsum(pca.explained_variance_ratio_) > ratio)[0][0]
    return pca.transform(data)[:, :i]


class FileKeyValueDataset:
    """Load datasets from HDF, torch, numpy or joblib files.

    The data is directly accessible through attributes of instances of this class.

    Args:
        path: The filepath for loading the data from. Should point to a file
            in a valid file format (hdf, torch, numpy, joblib). Valid extensions
            are ``jl``, ``joblib``, ``h5``, ``hdf``, ``hdf5``, ``pth``, ``pt`` and
            ``npz``.

    Example:

        >>> import cebra.io
        >>> import joblib
        >>> import tempfile
        >>> from pathlib import Path
        >>> tmp_file = Path(tempfile.gettempdir(), 'test.jl')
        >>> _ = joblib.dump({'foo' : 42}, tmp_file)
        >>> data = cebra.io.FileKeyValueDataset(tmp_file)
        >>> data.foo
        42

    """

    def __init__(self, path: str):
        self.path = path
        keys = []
        for key, value in self._iterate_items():
            setattr(self, key, value)
            keys.append(key)
        self.keys = tuple(keys)

    def __repr__(self):
        sizes = []
        for key in self.keys:
            val = getattr(self, key)
            val_str = type(val).__name__
            if isinstance(val, np.ndarray):
                val_str = f"{val_str}(shape={val.shape}, dtype={val.dtype})"
            sizes.append(f"{key} = {val_str}")
        sizes = ",\n  ".join(sizes)
        return f"{type(self).__name__}(keys=(\n  {sizes}\n))"

    def _iterate_items(self):
        extension = self.path.suffix
        if extension in [".jl", ".joblib"]:
            dataset = joblib.load(self.path)
        elif extension in [".h5", ".hdf", ".hdf5"]:
            raise NotImplementedError()
        elif extension in [".pth", ".pt"]:
            dataset = torch.load(self.path)
        elif extension in [".npz"]:
            dataset = np.load(self.path, allow_pickle=True)
        else:
            raise ValueError(f"Invalid file format: {extension} in {self.path}")
        for key in dataset.keys():
            yield key, dataset[key]
